@@ -28,16 +28,18 @@
 #' @importClassesFrom SummarizedExperiment SummarizedExperiment
 #' @importClassesFrom SingleCellExperiment SingleCellExperiment
 #' @importClassesFrom SpatialExperiment SpatialExperiment
-#' @importClassesFrom DelayedArray DelayedArray
-#' @importClassesFrom HDF5Array HDF5Array
+#' @importClassesFrom DelayedArray DelayedArray DelayedMatrix
+#' @importClassesFrom HDF5Array HDF5Array HDF5Matrix
+#' @importClassesFrom SparseArray SVT_SparseArray
 #'
 #' @aliases GsvaExprData
 #' @name GsvaExprData-class
 #' @rdname GsvaExprData-class
 #' @exportClass GsvaExprData
 setClassUnion("GsvaExprData",
-              c("matrix", "dgCMatrix", "ExpressionSet",
-                "SummarizedExperiment", "SingleCellExperiment", "SpatialExperiment", "HDF5Array"))
+              c("matrix", "dgCMatrix", "SVT_SparseArray", "DelayedMatrix",
+                "ExpressionSet", "SummarizedExperiment", "SingleCellExperiment",
+                "SpatialExperiment", "HDF5Matrix"))
 
 
 #' `GsvaGeneSets` class
@@ -328,6 +330,19 @@ setClass("ssgseaParam",
 #' to apply in the presence of missing values in the input expression data; see
 #' `ssgseaParam`.
 #'
+#' @slot nzcount Numeric vector of length 1. Number of non-zero values in the
+#' selected assay, if there is more than one, of the 'exprData' slot.
+#'
+#' @slot ondisk Character vector of length 1 denoting whether an on-disk backend
+#' should be used to reduce the memory footprint. The default value
+#' `ondisk="auto"` will attempt to load all the data in main memory when the
+#' number of nonzero values is equal or smaller than 2^31, otherwise it will
+#' attempt working with an on-disk data structure that reduces de memory
+#' footprint. When `ondisk="yes"` it will attempt to work with an on-disk data
+#' structure, while when `ondisk="no"` it will attempt to load all the data in
+#' main memory, irrespective of whether the number of nonzero values is larger,
+#' equal, or smaller than 2^31.
+#'
 #' @seealso
 #' [`GsvaExprData-class`],
 #' [`GsvaGeneSets-class`],
@@ -349,7 +364,9 @@ setClass("gsvaParam",
                  checkNA="character",
                  didCheckNA="logical",
                  anyNA="logical",
-                 use="character"),
+                 use="character",
+                 nzcount="numeric",
+                 ondisk="character"),
          contains="GsvaMethodParam",
          prototype=list(exprData=NULL,
                         geneSets=NULL,
@@ -366,7 +383,9 @@ setClass("gsvaParam",
                         checkNA=NA_character_,
                         didCheckNA=NA,
                         anyNA=NA,
-                        use=NA_character_))
+                        use=NA_character_,
+                        nzcount=NA_real_,
+                        ondisk=NA_character_))
 
 #' @name gsvaRanksParam-class
 #' @rdname gsvaParam-class
@@ -388,4 +407,6 @@ setClass("gsvaRanksParam",
                         checkNA=NA_character_,
                         didCheckNA=NA,
                         anyNA=NA,
-                        use=NA_character_))
+                        use=NA_character_,
+                        nzcount=NA_real_,
+                        ondisk=NA_character_))
