@@ -113,6 +113,20 @@
 #' happens, and giving an error if no values are left after removing the `NA`
 #' values.
 #'
+#' @param ondisk Character vector of length 1 denoting whether an on-disk backend
+#' should be used to reduce the memory footprint. The default value
+#' `ondisk="auto"` will attempt to load all the data in main memory when the
+#' number of nonzero values is equal or smaller than 2^31, otherwise it will
+#' attempt working with an on-disk data structure that reduces de memory
+#' footprint. When `ondisk="yes"` it will attempt to work with an on-disk data
+#' structure, while when `ondisk="no"` it will attempt to load all the data in
+#' main memory, irrespective of whether the number of nonzero values is larger,
+#' equal, or smaller than 2^31.
+#'
+#' @param verbose Logical vector of length 1. It gives information about some
+#' decisions made by the software during parameter object construction when
+#' `verbose=TRUE` (default) and remains silent otherwise.
+#'
 #' @return A new [`gsvaParam-class`] object.
 #'
 #' @seealso [`GeneIdentifierType`][GSEABase::GeneIdentifierType-class],
@@ -156,17 +170,21 @@ gsvaParam <- function(exprData, geneSets,
                       kcdfNoneMinSampleSize=200, tau=1, maxDiff=TRUE,
                       absRanking=FALSE, sparse=TRUE,
                       checkNA=c("auto", "yes", "no"),
-                      use=c("everything", "all.obs", "na.rm")) {
+                      use=c("everything", "all.obs", "na.rm"),
+                      ondisk=c("auto", "yes", "no"),
+                      verbose=TRUE) {
     kcdf <- match.arg(kcdf)
     kcdfNoneMinSampleSize <- as.integer(kcdfNoneMinSampleSize)
     checkNA <- match.arg(checkNA)
     use <- match.arg(use)
+    ondisk <- match.arg(ondisk)
 
     ## check assay parameter and assay names
-    assay <- .check_assayNames(assay, exprData)
+    assay <- .check_assayNames(assay, exprData, verbose)
 
     ## check for presence of valid row/feature names
-    exprData <- .check_rownames(exprData)
+    exprData <- .check_rowNames(expr=exprData, useDummyNames=TRUE,
+                                verbose=verbose)
 
     xa <- gsvaAnnotation(exprData)
     if(is.null(xa)) {
@@ -187,6 +205,17 @@ gsvaParam <- function(exprData, geneSets,
     naparam <- .check_for_na_values(exprData=exprData, assay=assay,
                                     checkNA=checkNA, use=use)
 
+    nzc <- .estimate_nzcount(exprData, assay, verbose)
+    if (ondisk == "auto") {
+      if (nzc >= .Machine$integer.max) {
+        ondisk <- "yes"
+      } else
+        ondisk <- "no"
+    }
+
+    if (!is_sparse(unwrapData(exprData, assay))) ## use sparse regime only
+        sparse <- FALSE                          ## when input is sparse
+
     new("gsvaParam",
         exprData=exprData, geneSets=geneSets,
         assay=assay, annotation=annotation,
@@ -194,7 +223,7 @@ gsvaParam <- function(exprData, geneSets,
         kcdf=kcdf, kcdfNoneMinSampleSize=kcdfNoneMinSampleSize,
         tau=as.double(tau), maxDiff=maxDiff, absRanking=absRanking,
         sparse=sparse, checkNA=checkNA, didCheckNA=naparam$didCheckNA,
-        anyNA=naparam$any_na, use=use)
+        anyNA=naparam$any_na, use=use, nzcount=nzc, ondisk=ondisk)
 }
 
 
@@ -291,6 +320,12 @@ setValidity("gsvaParam", function(object) {
     if(!.isCharLength1(object@use)) {
         inv <- c(inv, "@use must be a single character string")
     }
+    if(length(object@nzcount) != 1) {
+        inv <- c(inv, "@nzcount must be of length 1")
+    }
+    if(is.na(object@nzcount)) {
+        inv <- c(inv, "@nzcount must not be NA")
+    }
     return(if(length(inv) == 0) TRUE else inv)
 })
 
@@ -333,6 +368,12 @@ get_sparse <- function(object) {
   return(object@sparse)
 }
 
+#' @noRd
+get_ondisk <- function(object) {
+  stopifnot(inherits(object, "gsvaParam"))
+  return(object@ondisk)
+}
+
 ## getters for 'checkNA', 'didCheckNA' and 'use' are
 ## currently in ssgsea.R
 
@@ -347,6 +388,12 @@ setMethod("anyNA", signature=c("gsvaParam"),
           function(x, recursive=FALSE)
             return(x@anyNA))
 
+#' @importFrom SparseArray nzcount
+#' @aliases nzcount,gsvaParam-method
+setMethod("nzcount", signature=c("gsvaParam"),
+          function(x)
+            return(x@nzcount))
+
 
 ## ----- show -----
 
@@ -360,8 +407,7 @@ setMethod("show",
                   "maxDiff: ", get_maxDiff(object), "\n",
                   "absRanking: ", get_absRanking(object), "\n",
                   sep="")
-              if ("dgCMatrix" %in% class(unwrapData(get_exprData(object), get_assay(object))))
-                  cat("sparse: ", get_sparse(object), "\n")
+              cat("sparse: ", get_sparse(object), "\n")
               cat("checkNA: ", get_checkNA(object), "\n", sep="")
               if (get_didCheckNA(object)) {
                   if (anyNA(object)) {
@@ -371,6 +417,11 @@ setMethod("show",
                       cat("missing data: no\n")
               } else
                   cat("missing data: didn't check\n")
+              nzcmsg <- sprintf("nonzero values: %s than 2^31 (INT_MAX)\n",
+                                ifelse(nzcount(object) > .Machine$integer.max,
+                                       "more", "less"))
+              cat("ondisk: ", get_ondisk(object), "\n")
+              cat(nzcmsg)
           })
 
 ## ----- setters for gsvaRanksParam -----
